# Overview

This interface layer implements the basic form of the `rest` interface protocol,
which is used to communicate with services exposing a REST API requiring a password for
a trusted communication

# Usage

## Provides

By providing the `rest` interface, your charm is providing an HTTP server that
requires https and a password for clients to connect

Your charm needs to provide the port on which it is serving its content and the trust
password as soon as the `{relation_name}.available` state is set:

```python
@when('api.available')
def configure_api(api):
    api.configure(port=hookenv.config('port'), 
        trust_password=hookenv.config('trust_password'),
        node=hookenv.config('node_name'))
```

## Requires

By requiring the `rest` interface, your charm is consuming one or more HTTP
servers, as a trusted https REST endpoint.

Your charm should respond to the `{relation_name}.available` state, which
indicates that there is at least one HTTP server connected, offering a REST API.

The `services()` method returns a list of available HTTP services and their
associated hosts, ports, name and trust_password.

The return value is a list of dicts of the following form:

```python
[
    {
        'service_name': name_of_service,
        'hosts': [
            {
                'hostname': private_address_of_host,
                'public_address': public_address_of_host,
                'port': port_for_host,
                'trust_password': trust_password,
                'node': name_of_the_unit,
            },
            # ...
        ],
    },
    # ...
]
```

A trivial example of handling this interface would be:

```python
from charms.reactive.helpers import data_changed

@when('api.available')
def update_api_service_config(api):
    services = api.services()
    if not data_changed('api.services', services):
        return
    for service in services:
        for host in service['hosts']:
            hookenv.log('{} has a unit {}:{}'.format(
                service['service_name'],
                host['hostname'],
                host['public_address'],
                host['port'],
                host['trust_password'],
                host['node']))
```
