#
# Anbox - The Android in a Box runtime environment
# Copyright 2018 Canonical Ltd.  All rights reserved.
#

import os

from charmhelpers.core.hookenv import unit_private_ip
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from datetime import timedelta
from datetime import datetime

from OpenSSL.crypto import load_certificate, FILETYPE_PEM

import io
import ipaddress

from charms.layer.anbox import get_public_address


def create_cert_files_if_needed(cert_path, key_path):
    if (os.path.exists(cert_path) and os.path.exists(key_path)):
        return

    hostname = get_public_address()
    public_ip = hostname
    private_ip = unit_private_ip()
    cert, key = generate_selfsigned_cert(hostname, public_ip, private_ip)

    cert_base_path = os.path.dirname(cert_path)
    if not os.path.exists(cert_base_path):
        os.makedirs(cert_base_path, mode=0o0700)

    key_base_path = os.path.dirname(key_path)
    if not os.path.exists(key_base_path):
        os.makedirs(key_base_path, mode=0o0700)

    with open(os.open(cert_path, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
        f.write(str(cert, 'UTF-8'))

    with open(os.open(key_path, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
        f.write(str(key, 'UTF-8'))


def generate_selfsigned_cert(hostname, public_ip, private_ip):
    if not hostname:
        raise Exception('A hostname is required')

    if not public_ip:
        raise Exception('A public IP is required')

    if not private_ip:
        raise Exception('A private IP is required')

    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend()
    )

    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, hostname)
    ])
    alt_names = x509.SubjectAlternativeName([
        # best practice seem to be to include the hostname in the SAN
        x509.DNSName(hostname),
        # allow addressing by IP, for when you don't have real DNS
        # openssl wants DNSnames for ips...
        x509.DNSName(public_ip),
        x509.DNSName(private_ip),
        x509.IPAddress(ipaddress.IPv4Address(public_ip)),
        x509.IPAddress(ipaddress.IPv4Address(private_ip)),
    ])
    # path_len=0 means this cert can only sign itself, not other certs.
    basic_contraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(key.public_key())
            .serial_number(1000)
            .not_valid_before(now)
            .not_valid_after(now + timedelta(days=10*365))
            .add_extension(basic_contraints, False)
            .add_extension(alt_names, False)
            .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return cert_pem, key_pem


def read_cert(filepath):
    with open(filepath, 'r') as f:
        return f.read()


def cert_raw_content(cert):
    if not cert:
        return

    buff = io.StringIO(cert)
    content = ''

    line = buff.readline()
    while line and '-----BEGIN CERTIFICATE-----' not in line:
        line = buff.readline()

    line = buff.readline()
    while line and '-----END CERTIFICATE-----' not in line:
        content += line.rstrip('\r\n')
        line = buff.readline()

    return content


def fingerprint(cert):
    x509_cert = load_certificate(FILETYPE_PEM, cert)
    sha256 = x509_cert.digest("sha256")
    return str(sha256, 'UTF-8').replace(':', '').lower()
