#
# Copyright 2018 Canonical Ltd.  All rights reserved.
#

import filecmp
import json
import multiprocessing
import os
import re
import shutil
import socket
import subprocess
import tempfile
import time
from subprocess import (check_output, check_call, call,
                        CalledProcessError, TimeoutExpired, STDOUT)

import netifaces
import psutil

from charmhelpers import fetch
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.contrib.network.ip import get_host_ip
from charmhelpers.core import host, templating
from charmhelpers.core.hookenv import (config, log, open_port, close_port,
                                       status_set, local_unit, unit_private_ip,
                                       unit_public_ip, charm_name,
                                       resource_get, relation_ids, unit_get,
                                       storage_get, storage_list, application_version_set)
from charmhelpers.core.unitdata import kv
from charms import apt
from charms.layer import snap, lxd_socket
from charms.layer.anbox import (install_or_upgrade_anbox_package,
                                remove_anbox_package,
                                generate_random_string)
from charms.layer.pki import cert_raw_content, fingerprint
from charms.reactive import hook
from charms.reactive.decorators import when, when_not, when_any
from charms.reactive.flags import set_flag, clear_flag, is_flag_set
from charms.reactive.relations import endpoint_from_flag


class CertificateException(Exception):
    pass


LXD_READY_TIMEOUT = 600  # Seconds

SUBUID = '/etc/subuid'
SUBGID = '/etc/subgid'
ROOT_USER = 'root'
LXD_USER = 'lxd'

TRUST_PWD_LEN = 32

SECURITY_LIMITS_CONF = '/etc/security/limits.conf'
SYSCTL_CONF = '/etc/sysctl.conf'

UBUNTU_USER = 'ubuntu'

DKMS = 'dkms'
ANBOX_MODULES_DKMS = 'anbox-modules-dkms'

CLIENT_CERTIFICATE_KEY = 'client_certificate'
FINGERPRINT_KEY = 'fingerprint'

LXD_SNAP_COMMON = "/var/snap/lxd/common"
LXD_SNAP_DATA = "/var/snap/lxd/current"
LXD_SNAP_USER_COMMON = "/root/snap/lxd/common"
LXD_SNAP_USER_DATA = "/root/snap/lxd/current"

TELEGRAF_CONF_FOLDER = '/etc/telegraf/telegraf.d'
SOCKET_LISTENER_CONF = '{}/lxd_socket_listener.conf'.format(
    TELEGRAF_CONF_FOLDER)
CGROUP_CONF = '{}/cgroup.conf'.format(TELEGRAF_CONF_FOLDER)
SYSTEMD_SERVICE_FOLDER = '/etc/systemd/system'
TELEGRAF_LXD_WAIT_CONF = '{}/telegraf.service.d/amsbr0_wait.conf'.format(
    SYSTEMD_SERVICE_FOLDER)
AMS_BRIDGE_DEVICE = "amsbr0"
TELEGRAF_SERVICE = "telegraf.service"
WAIT_FOR_NETIF_PATH = "/usr/bin/wait-for-netif"

GPU_TYPE_NONE = 'none'
GPU_TYPE_NVIDIA = 'nvidia'
GPU_TYPE_AMD = 'amd'
GPU_TYPE_KEY = 'gpu_type'

FLAG_GPU_NVIDIA_INSTALLED = 'gpu.nvidia.installed'

DEFAULT_STORAGE_POOL = 'ams0'

SYSTEMD_SYSTEM_CONF_DIR = '/etc/systemd/system'
SYSTEMD_LXD_DROP_IN_FOLDER = '{}/snap.lxd.daemon.d'.format(
    SYSTEMD_SYSTEM_CONF_DIR)
SYSTEMD_LXD_DROP_IN_TWEAKS_CONF = '{}/tweaks.conf'.format(
    SYSTEMD_LXD_DROP_IN_FOLDER)

SYSTEMD_LXD_CONTAINERS_DROP_IN_FOLDER = '{}/lxd-containers.service.d'.format(
    SYSTEMD_SYSTEM_CONF_DIR)
SYSTEMD_LXD_CONTAINERS_IN_STOP_HACK_CONF = '{}/stop-hack.conf'.format(
    SYSTEMD_LXD_CONTAINERS_DROP_IN_FOLDER)

LXC_CMD_TIMEOUT = 300  # Seconds
LXD_CMD_TIMEOUT = LXD_READY_TIMEOUT

LXD_CLUSTERED_NODES_KEY = 'clustered_nodes'


def should_install_kmod():
    disabled = config('disable_kernel_modules') or False
    return not is_inside_lxd() and not disabled


def autodetect_pci_device(type, vendor):
    try:
        check_output(['/bin/sh', '-c',
                      "lspci | grep '{}' | grep -q {}".format(type, vendor)])
        return True
    except CalledProcessError:
        return False


def autodetect_gpu():
    if (autodetect_pci_device("VGA", "NVIDIA") or
            autodetect_pci_device("3D controller", "NVIDIA")):
        return GPU_TYPE_NVIDIA
    elif autodetect_pci_device("VGA", "AMD"):
        return GPU_TYPE_AMD
    return GPU_TYPE_NONE


def snap_set(param, value):
    """Sets a config parameter for this charm main snap"""
    try:
        check_output(
            ['snap', 'set', charm_name(), '{}={}'.format(param, value)],
            stderr=STDOUT)
    except CalledProcessError as ex:
        log('Could not set snap config param: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def enable_cpu_performance_mode():
    try:
        check_output(['systemctl', 'disable', 'ondemand'], stderr=STDOUT)
        check_output(['systemctl', 'mask', 'ondemand'], stderr=STDOUT)
    except CalledProcessError as ex:
        log('Failed to disable ondemand CPU governor: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def apply_lxd_containers_stop_hack():
    # HACK: This works around a problem with the lxd-container.service unit
    # hanging on a `lxd shutdown` when we perform the purge of the lxd packet
    # below. To avoid the hang we set the TimeoutStopSec option of the unit
    # to a very low value which will quickly terminate the command if it
    # doesn't come back in time

    if not os.path.exists(SYSTEMD_LXD_CONTAINERS_DROP_IN_FOLDER):
        os.mkdir(SYSTEMD_LXD_CONTAINERS_DROP_IN_FOLDER, mode=0o755)

    templating.render(
        source='lxd-containers-stop-hack.conf',
        target=SYSTEMD_LXD_CONTAINERS_IN_STOP_HACK_CONF,
        context={})

    check_call(['systemctl', 'daemon-reload'])


@hook('pool-storage-attached')
def pool_storage_attached():
    bootstrap()


@hook('pool-storage-detaching')
def pool_storage_detaching():
    purge_lxd()


@hook('install')
def install():
    bootstrap()


# Independent on which hook (install or pool-storage-attached) is called
# first we only want to run (and must run) the bootstrap as first thing
# but still only once. We use the bootstrap.done flag to prevent us
# running it multiple times.
def bootstrap():
    """
    Removes any previous deb version of lxd and prepares the
    machine for lxd configuration. That includes, configuring
    specific values for root into /etc/subuid and /etc/subgid,
    set security limits in /etc/security/limits.conf and
    modify /etc/sysctl.conf
    """

    if is_flag_set('bootstrap.done'):
        return

    apply_lxd_containers_stop_hack()

    # remove any previous deb version of lxd
    apt.purge(['lxd', 'lxd-client'])

    # prepare the machine for lxd
    configure_uid_mapping()
    set_security_limits()
    set_sysctl_config()
    setup_gpu_type(autodetect_gpu())
    setup_systemd_dropins()
    enable_cpu_performance_mode()

    check_and_update_kernel()

    if not is_inside_lxd():
        # We need ZFS utilities in order to destroy a storage pool manually
        # when it cannot be done by LXD itself
        apt.queue_install(['zfsutils-linux'])
        apt.install_queued()

    set_flag('bootstrap.done')


def set_charm_version():
    try:
        info = subprocess.run(['snap', 'info', 'lxd'],
                              stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as err:
        log('Failed to get snap info: {}'.format(err), level='ERROR')
    else:
        stdout = info.stdout.decode('utf8').splitlines()
        version = [line.split()[1] for line in stdout if 'installed' in line]
        if version:
            application_version_set(version[0])


def is_bionic_ga_kernel():
    try:
        check_output(['/bin/sh', '-c',
                      'uname -r | grep 4.15'])
        return True
    except CalledProcessError:
        return False


def is_aws_kernel():
    try:
        check_output(['/bin/sh', '-c',
                      'uname -r | grep -q aws'])
        return True
    except CalledProcessError:
        return False


def is_shiftfs_kernel_module_available():
    try:
        check_output(['/bin/sh', '-c',
                      'cat /proc/modules | grep -q shiftfs'])
        return True
    except CalledProcessError:
        return False


def set_unit_status():
    if not is_flag_set('lxd.clustered'):
        status_set('maintenance', 'Waiting for AMS to add node to the cluster')
    elif is_flag_set('reboot.required'):
        status_set('active', 'reboot required to activate new kernel')
    else:
        status_set('active', '')


def check_and_update_kernel():
    kernel_variant = get_kernel_variant()
    # We only need to update to a newer kernel on AWS with bionic
    if not is_bionic_ga_kernel():
        if not is_shiftfs_kernel_module_available():
            # Install the extra kernel modules package while shiftfs kernel
            # module is missing if the edge kernel is already installed. We
            # also need to install the meta package to pull in future new
            # kernel versions.
            kernel_module_extra_pkg = 'linux-modules-extra-{}'.format(get_kernel_release())
            pkgs = [kernel_module_extra_pkg]

            if is_meta_modules_extra_supported(kernel_variant):
                kernel_modules_extra_meta_pkg = 'linux-modules-extra-{}'.format(kernel_variant)
                pkgs.append(kernel_modules_extra_meta_pkg)

            fetch.apt_update()
            fetch.apt_install(pkgs)

            # Configure lxd to enable the shiftfs feature, which will reload
            # the shiftfs kernel module after restarting lxd snap without
            # rebooting a machine.
            configure_shiftfs(restart=True)

        return

    try:
        fetch.apt_update()
        kernel_meta_pkg = 'linux-{}-edge'.format(kernel_variant)
        if kernel_variant == 'generic':
            kernel_meta_pkg = 'linux-{}-edge'.format('generic-hwe-18.04')
        pkgs = [kernel_meta_pkg]

        if is_meta_modules_extra_supported(kernel_variant):
            kernel_modules_extra_meta_pkg = 'linux-modules-extra-{}-edge'.format(kernel_variant)
            pkgs.append(kernel_modules_extra_meta_pkg)

        fetch.apt_install(pkgs)
    except CalledProcessError as ex:
        raise ex

    set_flag('reboot.required')


@hook('stop')
def stop():
    """
    Removes LXD snap, cleans up unit local database and
    restores the values of /etc/sysctl.conf, /etc/security/limits.conf,
    /etc/subuid and /etc/subgid
    """
    remove_snap()
    uninstall_anbox_modules()
    kv().unset(FINGERPRINT_KEY)
    kv().flush()
    restore_uid_mapping()
    restore_security_limits()
    restore_sysctl_config()
    remove_systemd_dropins()
    setup_gpu_type(GPU_TYPE_NONE)


@when('upgrade.allowed')
def trigger_upgrade_charm():
    upgrade_charm()


def apply_lxd_debug_binary():
    lxd_debug_binary_path = os.path.join(LXD_SNAP_COMMON, 'lxd.debug')
    if not is_resource_attached('lxd-binary'):
        if os.path.exists(lxd_debug_binary_path):
            try:
                os.remove(lxd_debug_binary_path)
                check_call(['systemctl', 'reload', 'snap.lxd.daemon'])
            except OSError:
                pass
        return False

    lxd_binary_path = resource_get('lxd-binary')
    os.rename(lxd_binary_path, lxd_debug_binary_path)
    os.chmod(lxd_debug_binary_path, mode=0o0750)
    shutil.chown(lxd_debug_binary_path, user="root", group="root")
    check_call(['systemctl', 'reload', 'snap.lxd.daemon'])
    return True


@hook('upgrade-charm')
def upgrade_charm():
    # If only manual upgrades are allowed we wait for the upgrade.allowed
    # flag to be set before we proceed here on the hook execution
    only_manual_upgrades = config('enable_manual_upgrade') or False
    if only_manual_upgrades and not is_flag_set('upgrade.allowed'):
        log('Not running upgrade as automatic upgrades are disabled', level='INFO')
        return

    install_or_upgrade_anbox_package(ANBOX_MODULES_DKMS)

    if should_install_kmod():
        # The virt_wifi module is added in 1.11 and needs to be loaded on upgraded systems
        install_kernel_module('virt_wifi')

    gpu_type = kv().get(GPU_TYPE_KEY)
    installed_flag_name = "gpu.{}.installed".format(gpu_type)
    if is_flag_set(installed_flag_name):
        run_gpu_support_helper(gpu_type, "upgrade")

    if not apply_lxd_debug_binary():
        # Refresh LXD snap if an update is available
        snap.refresh(charm_name(), channel=config('channel'))
        set_charm_version()
        connect_lxd_plugs()

    # If we have an NRPE relation, force run of that as well
    if relation_ids('nrpe-exteral-master'):
        clear_flag('nagios.configured')

    set_security_limits()
    set_sysctl_config()

    clear_flag('upgrade.allowed')

    log('Finished charm upgrades', level='INFO')


@hook('update-status')
def update_status():
    """Checks if amsbr0 exists and updates lxd relation in such case"""
    if not exists_device(AMS_BRIDGE_DEVICE):
        return

    api = endpoint_from_flag('endpoint.api.available')
    if not api:
        log('Bridge exists but REST API still not available. Skip')
        return

    # Check telegraf status, because it could be deployed
    # after the node and relations set
    if os.path.exists(TELEGRAF_CONF_FOLDER):
        configure_telegraf_if_detected(
            [
                configure_telegraf_socket_listener,
                configure_telegraf_cgroups,
                configure_telegraf_amsbr0_wait
            ]
        )


@when('lxd.clustered')
def lxd_is_clustered():
    set_unit_status()


def process_and_update_api_info(api):
    node = normalize(local_unit())

    # Set up received client certificate
    for services in api.services():
        if CLIENT_CERTIFICATE_KEY in services.keys():
            client_certificate = services[CLIENT_CERTIFICATE_KEY]
            register_client_certificate(client_certificate)
        if LXD_CLUSTERED_NODES_KEY in services.keys():
            clustered_nodes = services[LXD_CLUSTERED_NODES_KEY]
            if node in clustered_nodes:
                set_flag('lxd.clustered')

    # Provide any required node data through the relation
    port = config('https_port')
    cpu_cores = num_cores()
    memory = available_memory()
    mtu = calculate_mtu()
    private_address = get_private_address()
    public_address = get_public_address()
    cert_fingerprints = get_registered_client_certificate_fingerprints()
    storage_device = get_storage_device()

    api.publish_info(port=port, node=node, cpu_cores=cpu_cores,
                     memory=memory, mtu=mtu,
                     private_address=private_address,
                     public_address=public_address,
                     cert_fingerprints=cert_fingerprints,
                     storage_device=storage_device)


@when('endpoint.api.available', 'snap.installed.lxd')
@when_not('api.configured')
def endpoint_api_available(api):
    """LXD can expose its REST API because a relation has been set"""
    configure_rest_remote_access()
    process_and_update_api_info(api)
    set_flag('api.configured')


@when('endpoint.api.changed', 'snap.installed.lxd')
def endpoint_api_changed(api):
    process_and_update_api_info(api)
    clear_flag('endpoint.api.changed')


@when('endpoint.api.available', 'snap.installed.lxd',
      'config.changed.telegraf_socket_listener_port')
def telegraf_socket_listener_port_changed(api):
    """ Reconfigures telegraf socket listener"""
    configure_telegraf_if_detected([configure_telegraf_socket_listener])


@when('endpoint.api.available', 'snap.installed.lxd',
      'config.changed.channel')
def channel_changed(api):
    """Reconfigures telegraf cgroups if needed"""
    status_set('maintenance', 'Installing LXD snap from {} channel'.format(
                               config('channel')))

    snap.refresh(charm_name(), channel=config('channel'))
    set_charm_version()
    wait_for_lxd_ready()
    configure_telegraf_if_detected([configure_telegraf_cgroups])

    set_unit_status()


@when('config.changed.shiftfs_enabled', 'snap.installed.lxd')
def shiftfs_enabled_changed():
    wait_for_lxd_ready()
    configure_shiftfs(restart=True)


@when('config.changed.debug', 'snap.installed.lxd')
def debug_changed():
    wait_for_lxd_ready()
    configure_debug(restart=True)


@hook('api-relation-broken')
def api_relation_broken(*args):
    """
    Clean up node and leave everything pristine just in case
    this node is added back to the cluster
    """
    purge_lxd()

    # If we were marked as clustered before we are now not clustered anymore
    clear_flag('lxd.clustered')

    # We need to configure the REST API access again as we have purged
    # all configuration
    clear_flag('api.configured')

    set_unit_status()


# TODO We cannot remove the certificate from trusted ones. As we are
# in a cluster, maybe other node wants to use that same certificate
# as a valid one for the client operations.
# We have to figure out a way to do this when last node is removed.
# That information is not possible to obtain here and AMS cannot
# probably send anything to the last lxd node when it is departed
# though let's see
#
# @hook('api-relation-broken')
# def api_relation_broken(*args):
#     fingerprint = kv().get(FINGERPRINT_KEY, default=None)
#     if fingerprint:
#         unregister_client_certificate(fingerprint)


def configure_shiftfs(restart):
    if not is_flag_set('snap.installed.lxd'):
        return
    shiftfs_enabled = config('shiftfs_enabled')
    snap_set('shiftfs.enable', "{}".format(shiftfs_enabled).lower())
    if restart:
        restart_snap()


def configure_debug(restart):
    debug = config('debug')
    snap_set('daemon.debug', "{}".format(debug).lower())
    if restart:
        restart_snap()


def wait_for_snapd_seeded():
    check_output(['snap', 'wait', 'system', 'seed.loaded'])


def wait_for_snapd():
    """
    This waits for snapd to be ready to install any new snaps. When
    we bootstrap on a fresh created cloud instance we may not have
    all latest snaps install and may see interferences of snapd
    updating itself during out own installation process. To avoid
    this we ensure all existing snaps are refreshed and the system
    is fully seeded. Afterwards its safe for us to perform our own
    installation.
    """
    wait_for_snapd_seeded()

    # We don't fail here when the snap refresh fails as network connectivity
    # might not be given. Instead we log the error and proceed.
    try:
        check_output(['snap', 'refresh'])
    except CalledProcessError as ex:
        log('Failed to refresh system snaps: {}'.format(str(ex.output, 'utf-8')),
            level='ERROR')

    wait_for_snapd_seeded()


@when('config.changed.images_compression_algorithm', 'snap.installed.lxd')
def configure_lxd():
    try:
        compression_algorithm = config('images_compression_algorithm')
        lxc('config', 'set', 'images.compression_algorithm', compression_algorithm)
    except CalledProcessError as ex:
        raise ex


def connect_lxd_plugs():
    # If the LXD snap is coming from a resource, we have to connect additional
    # plugs here to allow LXD to work
    if is_resource_attached('lxd'):
        snap.connect('lxd:lxd-support', 'core')
        snap.connect('lxd:system-observe', 'core')
        check_call(['snap', 'alias', 'lxd.lxc', 'lxc'])


def is_snap_installed(name):
    try:
        out = check_output(["snap", "info", name]).decode('utf-8')
    except subprocess.CalledProcessError:
        return False
    return "installed:" in out


@when('config.changed.channel')
@when_not('snap.installed.lxd')
def install_snap():
    """
    Installs lxd snap, creates lxd system group and waits for LXD to
    be available to activate the service
    """
    os.system('newgrp {}'.format(charm_name()))
    add_user_to_group(UBUNTU_USER, charm_name())
    wait_for_snapd()

    snap_name = charm_name()
    snap_channel = config('channel')

    # In case that the LXD snap is already installed we need to reinstall here
    # as the snap.install call with fail silently otherwise. On focal the LXD
    # snap is preinstalled from the 4.0/stable channel so we run into this on
    # every focal based installation.
    if is_snap_installed(snap_name):
        # We have to purge the snap here as layer-snap needs a call to
        # snap.install to put relevant flags in place we may miss otherwise
        check_call(['snap', 'remove', '--purge', 'lxd'])

    snap.install(snap_name, channel=snap_channel)

    set_charm_version()
    connect_lxd_plugs()
    configure_shiftfs(restart=False)
    configure_debug(restart=True)
    wait_for_lxd_ready()
    wait_for_snapd()
    apply_lxd_debug_binary()
    configure_lxd()
    configure_rest_remote_access(wait=False)
    close_old_and_open_new_ports('https_port')
    set_unit_status()


def close_old_and_open_new_ports(port_parameter):
    """Close the previous port and open the port from configuration."""
    configuration = config()
    previous_port = configuration.previous(port_parameter)
    port = configuration.get(port_parameter)
    if previous_port is not None and previous_port != port:
        log('The port changed; closing {0} opening {1}'.format(
            previous_port, port))
        close_port(previous_port)
    open_port(port)


def configure_rest_remote_access(wait=True):
    """
    If not already configured, sets host and port to listen, and
    trust password to use for clients using the exposed REST API
    """

    if wait:
        wait_for_lxd_ready()

    if is_rest_already_configured_in_the_cluster():
        return

    host = get_private_address()
    port = int(config('https_port'))
    password = generate_random_string(TRUST_PWD_LEN)

    if not port:
        port = 8443

    lxc('config', 'set', 'core.https_address', '{}:{}'.format(host, port))
    lxc('config', 'set', 'core.trust_password', password)


def is_rest_already_configured_in_the_cluster():
    """Returns true if there is a configured trust password for the REST API"""
    output = lxc('config', 'get', 'core.trust_password')
    return 'true' in output


def remove_snap():
    """Removes lxd snap and lxd system group"""
    snap.remove(charm_name())
    remove_user_from_group(UBUNTU_USER, charm_name())
    os.system('groupdel {}'.format(charm_name()))


def available_memory():
    """Returns the available memory in Gigabytes in current machine"""
    mem_in_gb = float(psutil.virtual_memory().total)/(1024*1024*1024)
    if mem_in_gb > 0:
        return '{}GB'.format(int(mem_in_gb))

    mem_in_mb = float(psutil.virtual_memory().total)/(1024*1024)
    return '{}MB'.format(int(mem_in_mb))


def num_cores():
    """Returns the available number of cores in current machine"""
    return multiprocessing.cpu_count()


def calculate_mtu():
    private_address = get_private_address()
    dev = get_dev_for_ip(private_address)
    return get_mtu_for_device(dev)


def get_dev_for_ip(ip):
    for dev in netifaces.interfaces():
        addresses = netifaces.ifaddresses(dev)
        if netifaces.AF_INET not in addresses:
            continue

        if 'addr' not in addresses[netifaces.AF_INET][0]:
            continue

        dev_ip = addresses[netifaces.AF_INET][0]['addr']
        if dev_ip == ip:
            return dev
    raise Exception('Not found a device for the requested IP')


def get_kernel_release():
    try:
        output = check_output(['uname', '-r'], stderr=STDOUT)
        return str(output, 'utf-8').strip()
    except CalledProcessError as ex:
        log('Error getting kernel release info: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def get_kernel_variant():
    try:
        output = check_output(['/bin/sh', '-c',
                               'uname -r | cut -d- -f3'])
        return str(output, 'utf-8').strip()
    except CalledProcessError as ex:
        log('Error getting kernel variant info: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def get_mtu_for_device(dev):
    '''Returns the MTU defined for a network device'''
    # WORKAROUND: Special case for packet.net deployments.
    # In such cases we detect, for example, bond0:0 dev,
    # but we really need to obtain the MTU from bond0.
    # To solve it, let's remove any : found in the name of
    # the device and check MTU in the main network interface
    dotIndex = dev.find(':')
    if dotIndex != -1:
        dev = dev[:dotIndex]

    mtuinfo = check_output(['ip', 'link', 'show', str(dev)])
    mtu = re.search('.*mtu ([0-9]+) .*', str(mtuinfo)).groups()[0]
    return int(mtu)


def get_ip_for_dev(dev):
    addresses = netifaces.ifaddresses(dev)
    if (netifaces.AF_INET not in addresses or
            'addr' not in addresses[netifaces.AF_INET][0]):
        raise Exception('Not found an IP for the requested device')

    return addresses[netifaces.AF_INET][0]['addr']


def exists_device(dev):
    try:
        addr = netifaces.ifaddresses(dev)
        return netifaces.AF_INET in addr
    except Exception:
        return False


def get_private_address():
    private_address = config('https_host')
    if not private_address:
        # Working around LP #1933303; When deployed inside LXD Juju will give us
        # the address of the amsbr0 network device if present rather than the
        # address of the eth0 device as result for unit_private_ip. As using the
        # incorrect address (which might not be available anymore in case that
        # the amsbr0 bridge is gone) will cause the `lxc config set ...` call
        # below to fail, we statically determine the address of the eth0 network
        # device here and use it as host address when we're running inside a
        # LXD container
        if is_inside_lxd():
            private_address = get_ip_for_dev("eth0")
        else:
            private_address = get_ip_address(unit_private_ip())
    return private_address


def get_public_address():
    public_address = get_ip_address(unit_public_ip())
    public_interface = config('public_interface')
    if public_interface:
        public_address = get_ip_for_dev(public_interface)
        if not public_address:
            log('Could not obtain a valid IP for the configured public_interface. \
                Using the default one')
            public_address = get_ip_address(unit_public_ip())
    return public_address


def get_ip_address(address):
    if is_ip_address(address):
        return address

    for interface in netifaces.interfaces():
        if interface != 'lo':
            i = netifaces.ifaddresses(interface)
            return i[netifaces.AF_INET][0]['addr']

    raise Exception('Unable to find default network interface')


def is_ip_address(address):
    try:
        socket.inet_aton(address)
        return True
    except socket.error:
        return False


def is_inside_lxd():
    return os.path.exists("/dev/lxd")


def get_storage_device():
    # We can't use any other storage device than dir inside a LXD
    # container so we force that here unless explicitly overriden
    # from the user by setting it in the Juju configuration
    if is_inside_lxd():
        return "dir"

    storage_ids = storage_list('pool')
    # Technically we specify in the metadata that the storage we
    # support is a singleton but for sanity we check here for more
    # and then only use the first one in the list
    if len(storage_ids) >= 1:
        return storage_get('location', storage_ids[0])

    # Choose 15 GB < x < 100GB, where x is 20% of the disk size. LXD applies
    # the same rule when creating a storage device. This should give us a
    # enough disk space for experiments. For a production deployment a
    # dedicated disk device or ZFS pool should be configured.
    usage = shutil.disk_usage("/var/snap/lxd/common")
    storage_device_size = (usage.total * 0.6) // (2**30)
    if storage_device_size > 100:
        storage_device_size = 100
    if storage_device_size < 15:
        storage_device_size = 15
    return "%dGB" % storage_device_size


def configure_uid_mapping():
    """Extend root user /etc/{subuid,subgid} mapping for LXD use"""
    subuid_backup = '{}.bak'.format(SUBUID)
    subgid_backup = '{}.bak'.format(SUBGID)
    if os.path.isfile(SUBUID):
        shutil.copy(SUBUID, subuid_backup)
    if os.path.isfile(SUBGID):
        shutil.copy(SUBGID, subgid_backup)

    for uidfile in (SUBUID, SUBGID):
        with open(uidfile, 'r+') as f_id:
            ids = []
            for s_id in f_id.readlines():
                _id = s_id.strip().split(':')
                if (_id[0] == ROOT_USER or _id[0] == LXD_USER):
                    _id[1] = 100000
                    _id[2] = 1000000
                ids.append(_id)
            f_id.seek(0)
            for _id in ids:
                f_id.write('{}:{}:{}\n'.format(*_id))
            f_id.truncate()


def restore_uid_mapping():
    """Restore original /etc/{subuid, subgid} mapping"""
    subuid_backup = '{}.bak'.format(SUBUID)
    subgid_backup = '{}.bak'.format(SUBGID)
    if os.path.isfile(subuid_backup):
        os.rename(subuid_backup, SUBUID)
    if os.path.isfile(subgid_backup):
        os.rename(subgid_backup, SUBGID)


def set_security_limits():
    """
    Creates /etc/security/limits.conf file based on a template
    NOTE: we don't want to override the original backup security
          limits conf file on charm upgrade.
    """
    backup = '{}.bak'.format(SECURITY_LIMITS_CONF)
    if not os.path.isfile(backup) or os.path.isfile(SECURITY_LIMITS_CONF):
        shutil.copy(SECURITY_LIMITS_CONF, backup)

    templating.render(
        source='security_limits.conf', target=SECURITY_LIMITS_CONF, context={})


def restore_security_limits():
    """Restores original /etc/security/limits.conf file"""
    backup = '{}.bak'.format(SECURITY_LIMITS_CONF)
    if os.path.isfile(backup):
        os.rename(backup, SECURITY_LIMITS_CONF)


def set_sysctl_config():
    """
       Creates /etc/sysctl.conf file based on a template
       NOTE: we don't want to override the original backup sysctl
             conf file on charm upgrade.
    """
    backup = '{}.bak'.format(SYSCTL_CONF)
    if not os.path.isfile(backup) and os.path.isfile(SYSCTL_CONF):
        shutil.copy(SYSCTL_CONF, backup)

    templating.render(source='sysctl.conf', target=SYSCTL_CONF, context={})

    # We have to explicitly apply the new settings to get them live on the system
    if not is_inside_lxd():
        check_call(['sysctl', '-p', SYSCTL_CONF])


def restore_sysctl_config():
    """Restores original /etc/sysctl.conf file"""
    backup = '{}.bak'.format(SYSCTL_CONF)
    if os.path.isfile(backup):
        os.rename(backup, SYSCTL_CONF)


def restart_snap():
    try:
        check_output(['systemctl', 'stop', 'snap.lxd.daemon'], stderr=STDOUT)
        check_output(['systemctl', 'start', 'snap.lxd.daemon'], stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error when restarting LXD snap: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def wait_for_lxd_ready():
    """Blocks current thread until lxd snap is ready"""
    lxd('waitready', '--timeout', str(LXD_READY_TIMEOUT))


def lxc(*args):
    """Wrapper for 'lxc' command"""
    params = ['/snap/bin/lxc']
    for it in args:
        params.append(it)
    s = ' '
    cmd = s.join(params)
    try:
        out = check_output(cmd, shell=True, stderr=STDOUT, timeout=LXC_CMD_TIMEOUT)
        return str(out, 'UTF-8')
    except TimeoutExpired as ex:
        log('Timed out for LXC command: `{}`: {} - {}'.format(
                ex.cmd, ex.timeout, str(ex.output, 'utf-8')))
        raise ex
    except CalledProcessError as ex:
        log('Error executing `{}`: {} - {}'.format(ex.cmd, ex.returncode,
                                                   str(ex.output, 'utf-8')))
        raise ex


def lxd(*args):
    """Wrapper for 'lxd' command"""
    params = ['/snap/bin/lxd']
    for it in args:
        params.append(it)
    return exec_cmd(params, LXD_CMD_TIMEOUT)


def exec_cmd(cmd, timeout=None):
    try:
        out = check_output(cmd, stderr=STDOUT, timeout=timeout)
        return str(out, 'UTF-8')
    except CalledProcessError as ex:
        log('Error executing `{}`: {} - {}'.format(ex.cmd, ex.returncode,
                                                   str(ex.output, 'utf-8')))
        raise ex


def normalize(name):
    """Removes not allowed '/' symbol"""
    return name.replace('/', '')


def add_user_to_group(usr, grp):
    """Adds user to an existing system group"""
    os.system('usermod -a -G {} {}'.format(grp, usr))


def remove_user_from_group(usr, grp):
    """Removes a user from a system group"""
    os.system(' gpasswd -d {} {}'.format(usr, grp))


def is_resource_attached(name):
    """
    Used to fetch the resource path of the given name.
    This wrapper obtains a resource path and adds an additional
    check to return False if the resource is zero length.
    """
    res_path = resource_get(name)
    if not res_path or os.stat(res_path).st_size == 0:
        return False
    return len(res_path) > 0


def get_registered_client_certificate_fingerprints():
    """
    Returns the array of fingerprints of the registered
    client certificates by querying the REST API
    """

    code, result = lxd_socket.get('1.0/certificates')
    if code >= 400:
        raise CertificateException(result)

    data = json.loads(result)
    resources = data['metadata']

    fingerprints = []
    for resource in resources:
        fingerprint = os.path.basename(resource.strip('\'"'))
        fingerprints.append(fingerprint)

    return fingerprints


def register_client_certificate(cert):
    """
    Registers a client certificate through LXD REST interface
    Certificate fingerprint is stored in kv to be used when removing
    the certificate from the trusted list if the relation is broken
    """
    raw_content = cert_raw_content(cert)
    if not raw_content:
        raise CertificateException('Could not get certificate raw content')

    payload = {
        "type": "client",
        "certificate": raw_content
    }
    result_code, result_text = lxd_socket.post('1.0/certificates',
                                               json.dumps(payload))

    if 200 <= result_code < 300:
        log('Registered client certificate: {}'.format(cert))
        # Save cert fingerprint in kv
        key_fingerprint = fingerprint(cert)
        if not key_fingerprint:
            raise CertificateException('Could not get certificate fingerprint')
        kv().set(FINGERPRINT_KEY, key_fingerprint)
        return

    if "Certificate already in trust store" in result_text:
        log('Certificate already provisioned. Skip provision')
    else:
        raise CertificateException(result_text)


def unregister_client_certificate(fingerprint):
    """Unregisters a client certificate through LXD REST interface"""
    if not fingerprint:
        return
    result_code, _ = lxd_socket.delete(
        '1.0/certificates/{}'.format(fingerprint))
    if result_code >= 200 and result_code < 300:
        log('Removed client certificate with fingerprint: {}'
            .format(fingerprint))
        return
    log('Warn: Could not unregister certificate')


def is_meta_modules_extra_supported(kernel_variant):
    meta_pkg_supported_kernel = ["aws", "aws-edge",
                                 "azure", "azure-edge",
                                 "gcp", "gcp-edge"]
    return kernel_variant in meta_pkg_supported_kernel


def install_anbox_modules_deps():
    # The `binder_linux` and `ashmem_linux` kernel modules are
    # dropped from anbox-modules-dkms package since both are available
    # on bionic and focal through the 5.4 kernel extra modules package.
    kernel_variant = get_kernel_variant()
    kernel_headers_pkg = 'linux-headers-{}'.format(get_kernel_release())
    kernel_headers_meta_pkg = 'linux-headers-{}'.format(kernel_variant)
    kernel_module_extra_pkg = 'linux-modules-extra-{}'.format(get_kernel_release())
    pkgs = [DKMS, kernel_headers_meta_pkg, kernel_headers_pkg, kernel_module_extra_pkg]

    if is_meta_modules_extra_supported(kernel_variant):
        kernel_modules_extra_meta_pkg = 'linux-modules-extra-{}'.format(kernel_variant)
        pkgs.append(kernel_modules_extra_meta_pkg)
    elif kernel_variant == "generic":
        # The generic kernel has (for historical reasons) a different meta package
        # to track linux-modules-extra
        pkgs.append("linux-image-generic")

    apt.queue_install(pkgs)
    apt.install_queued()


def uninstall_anbox_modules_deps():
    kernel_variant = get_kernel_variant()
    kernel_headers_pkg = 'linux-headers-{}'.format(get_kernel_release())
    kernel_module_extra_pkg = 'linux-modules-extra-{}'.format(get_kernel_release())
    kernel_headers_meta_pkg = 'linux-headers-{}'.format(kernel_variant)
    pkgs = [kernel_headers_pkg, kernel_headers_meta_pkg, DKMS, kernel_module_extra_pkg]
    if is_meta_modules_extra_supported(kernel_variant):
        kernel_modules_extra_meta_pkg = 'linux-modules-extra-{}'.format(kernel_variant)
        pkgs.append(kernel_modules_extra_meta_pkg)
    fetch.apt_purge(pkgs)


@when('ua.ready')
@when_not('anbox-modules.installed')
def install_anbox_modules():
    """Installs anbox modules read as a debian package resource"""

    if not should_install_kmod():
        log('Not installing kernel modules')
        set_flag("anbox-modules.installed")
        if is_flag_set('snap.installed.lxd'):
            set_unit_status()
        return

    install_anbox_modules_deps()

    install_or_upgrade_anbox_package(ANBOX_MODULES_DKMS)

    install_kernel_module('binder_linux')
    install_kernel_module('ashmem_linux')
    install_kernel_module('anbox_sync')
    install_kernel_module('virt_wifi')

    log('{} installed ok.'.format(ANBOX_MODULES_DKMS))
    set_flag("anbox-modules.installed")

    if is_flag_set('snap.installed.lxd'):
        set_unit_status()


def uninstall_anbox_modules():
    """Removes anbox-modules-dkms package from the system if installed"""
    if not is_flag_set('anbox-modules.installed'):
        return

    remove_anbox_package(ANBOX_MODULES_DKMS)

    remove_kernel_module('ashmem_linux')
    # NOTE: The binder kernel module cannot be unloaded at runtime
    # and the machine needs to be rebooted
    # remove_kernel_module('binder_linux')
    remove_kernel_module('anbox_sync')
    uninstall_anbox_modules_deps()


def snap_exists(name):
    check_exists_cmd = 'snap info  {}'.format(name)
    out = call(check_exists_cmd, shell=True)
    return out == 0


def install_kernel_module(module):
    try:
        check_call(['modprobe', module])
    except CalledProcessError as ex:
        log('Error installing {} kernel module: {}'
            .format(module, ex.returncode))
        raise ex


def remove_kernel_module(module):
    try:
        check_call(['rmmod', module])
    except CalledProcessError as ex:
        log('Error removing {} kernel module: {}'
            .format(module, ex.returncode))
        # Note: we don't rethrow the exception here when something goes
        # wrong with kernel module uninstallation as the error is not
        # critical while the unit is stopping


def count_lxd_nodes_in_cluster():
    output = lxc('info')
    if "server_clustered: false" in output:
        log('LXD is not clustered')
        return 1

    output = lxc('cluster', 'ls', '--format=json')
    nodes = json.loads(output)
    return len(nodes)


def purge_lxd():
    """
    Cleans up LXD data and leaves node ready to be added
    back to the cluster if needed
    """

    name = normalize(local_unit())
    num_retries = 30
    for n in range(num_retries):
        try:
            if count_lxd_nodes_in_cluster() > 1:
                log("Removing node {} from cluster".format(name))
                lxc('cluster', 'remove', name)
            else:
                log("Node {} is the last one in the cluster".format(name))

            # No error occured, we're done and can proceed as normal
            break

        except CalledProcessError as ex:
            # In some cases LXD reports an error when we try to remove
            # a node but another add or remove operation is already in
            # progress. To avoid a direct failure here and the charm
            # hook execution to be marked as failed, we detect the
            # error and retry a few times before we give up
            err = str(ex.output, 'utf-8')
            if ("configuration change is already in progress" in err and
                    n < num_retries - 1):
                log('LXD busy with cluster configuration changes, retrying')
                time.sleep(10)
                continue

            # If the node was already removed in the mean time we can stop
            # here and don't need to continue trying
            if "not found" in err:
                break

            if ("Node is not clustered" in err or
                    "LXD server isn't part of a cluster" in err):
                break

            log('Failed to remove node {} from cluster: {}'.format(
                name, err))
            raise ex

    snap.disable(charm_name())

    if not is_inside_lxd():
        # Deleting the pool via LXD sometimes failed because some container
        # would still be using it. We can't disable LXD and then ask it to
        # remove the storage pool, so we use zfs-utils directly to do it
        try:
            check_call(['zpool', 'list', DEFAULT_STORAGE_POOL])
        except CalledProcessError as e:
            if 'not found' not in str(e) or 'no such pool' not in str(e):
                log('Failed to get {} zfs pool: {}'.format(DEFAULT_STORAGE_POOL, e), level='ERROR')
        else:
            log('Removing {} zfs pool'.format(DEFAULT_STORAGE_POOL), level='INFO')
            check_call(['zpool', 'destroy', '-f', DEFAULT_STORAGE_POOL])

    purge_folder(LXD_SNAP_COMMON)
    purge_folder(LXD_SNAP_DATA)
    purge_folder(LXD_SNAP_USER_COMMON)
    purge_folder(LXD_SNAP_USER_DATA)
    snap.enable(charm_name())
    wait_for_lxd_ready()


def purge_folder(folder_path):
    if not os.path.exists(folder_path):
        return

    for file_object in os.listdir(folder_path):
        file_object_path = os.path.join(folder_path, file_object)
        if os.path.isfile(file_object_path):
            os.unlink(file_object_path)
        else:
            if not os.path.islink(file_object_path):
                shutil.rmtree(file_object_path, ignore_errors=True)


def configure_telegraf_if_detected(fns):
    """
    If detected that the default config path is available
    for telegraf, configure the specified plugins
    """
    if (not os.path.exists(TELEGRAF_CONF_FOLDER) or
            not os.path.isdir(TELEGRAF_CONF_FOLDER)):
        return

    restart = False
    for fn in fns:
        restart |= fn()

    if restart:
        log('Restarting telegraf service after applying configuration changes')
        host.service_restart(TELEGRAF_SERVICE)


def configure_telegraf_socket_listener():
    """
    Configure the socket listener listening for incoming tcp
    requests on amsbr0 device and return true if telegraf
    restart is needed, otherwise return false
    """
    if not exists_device(AMS_BRIDGE_DEVICE):
        return False

    # If configured port is already in use and socket listener file
    # was not changed, then socket is already configured, so skip
    address = get_ip_for_dev(AMS_BRIDGE_DEVICE)
    port = config('telegraf_socket_listener_port')
    if (port_in_use(port) and
            not telegraf_socket_file_changed(address, port)):
        return False

    templating.render(
        source='telegraf_socket_listener.conf',
        target=SOCKET_LISTENER_CONF,
        context={
            'ip': address,
            'port': port
        })

    close_old_and_open_new_ports('telegraf_socket_listener_port')

    return True


def configure_telegraf_cgroups():
    """
    Setup telegraf to report cgroup metrics and return true if
    telegraf restart is needed, otherwise return false
    """
    # Depending on the lxd charm version, the cgroups are in 'lxc' or
    # 'lxc.payload.<name>' group path
    channel = config('channel')
    group = 'lxc.payload*'
    if '3.0/stable' in channel:
        group = 'lxc'

    # We won't override the existing cgroup configuration file if the
    # cgroup configuration file is not changed. If so, we won't restart
    # the telegraf either.
    if (os.path.exists(CGROUP_CONF) and
            not cgroup_configuration_changed(group)):
        return False

    templating.render(
        source='telegraf_cgroup.conf',
        target=CGROUP_CONF,
        context={'group': group}
    )

    return True


def configure_telegraf_amsbr0_wait():
    """
    Configure telegraf to wait for amsbr0 to appear on startup
    """
    restart_necessary = False

    if not os.path.exists(WAIT_FOR_NETIF_PATH):
        templating.render(
            source='wait-for-netif.sh',
            target=WAIT_FOR_NETIF_PATH,
            context={}
        )
        check_call(['chmod', '+x', WAIT_FOR_NETIF_PATH])
        restart_necessary = True

    if not os.path.exists(TELEGRAF_LXD_WAIT_CONF):
        templating.render(
            source='telegraf_lxd_wait.conf',
            target=TELEGRAF_LXD_WAIT_CONF,
            context={'bridge': AMS_BRIDGE_DEVICE}
        )
        check_call(['systemctl', 'daemon-reload'])
        restart_necessary = True

    return restart_necessary


def port_in_use(port):
    cmd = "lsof -Pi :%s -sTCP:LISTEN > /dev/null" % port
    out = call(cmd, shell=True)
    return out == 0


def cgroup_configuration_changed(group):
    """
    Check if the cgroup configuration file was changed
    """
    if not os.path.exists(CGROUP_CONF):
        return True

    with tempfile.NamedTemporaryFile(suffix='.conf') as f:
        templating.render(
            source='telegraf_cgroup.conf',
            target=f.name,
            context={'group': group}
        )

        # filecmp.cmp uses a cache for past comparisons and the results,
        # We need to clear the entire cache before comparing both files
        filecmp.clear_cache()

        return not filecmp.cmp(f.name, CGROUP_CONF, shallow=False)


def telegraf_socket_file_changed(address, port):
    """
    Check if the telegraf socket file was changed
    """
    if not os.path.exists(SOCKET_LISTENER_CONF):
        return True

    with tempfile.NamedTemporaryFile(suffix='.conf') as f:
        templating.render(
            source='telegraf_socket_listener.conf',
            target=f.name,
            context={
                'ip': address,
                'port': port
            })

        # filecmp.cmp uses a cache for past comparisons and the results,
        # We need to clear the entire cache before comparing both files
        filecmp.clear_cache()

        return not filecmp.cmp(f.name, SOCKET_LISTENER_CONF, shallow=False)


def run_gpu_support_helper(gpu_type, action):
    prebuilt_kernel_modules = config(
        'use_prebuilt_nvidia_kernel_modules') or False
    try:
        cmd = ['scripts/gpu-support.sh', gpu_type, action]
        if prebuilt_kernel_modules:
            cmd.extend(['--prebuilt-kernel-modules'])
        check_call(cmd)
    except CalledProcessError as ex:
        log('Failed to {} {} GPU driver: {}'
            .format(action, gpu_type, ex.returncode))
        raise ex


def has_gpu_support(type):
    installed_flag_name = "gpu.{}.installed".format(type)
    return is_flag_set(installed_flag_name)


def add_gpu_support(type):
    installed_flag_name = "gpu.{}.installed".format(type)
    run_gpu_support_helper(type, "install")
    set_flag(installed_flag_name)
    kv().set(GPU_TYPE_KEY, type)


def remove_current_gpu_support():
    old_gpu_type = kv().get(GPU_TYPE_KEY)
    installed_flag_name = "gpu.{}.installed".format(old_gpu_type)
    if is_flag_set(installed_flag_name):
        run_gpu_support_helper(old_gpu_type, "uninstall")
        clear_flag(installed_flag_name)

    kv().set(GPU_TYPE_KEY, GPU_TYPE_NONE)


def setup_gpu_type(gpu_type):
    if gpu_type == GPU_TYPE_NONE:
        remove_current_gpu_support()
        return

    if not has_gpu_support(gpu_type):
        remove_current_gpu_support()
        add_gpu_support(gpu_type)


def setup_systemd_dropins():
    if not os.path.exists(SYSTEMD_LXD_DROP_IN_FOLDER):
        os.mkdir(SYSTEMD_LXD_DROP_IN_FOLDER, mode=0o755)

    templating.render(
        source='lxd-service-tweaks.conf',
        target=SYSTEMD_LXD_DROP_IN_TWEAKS_CONF,
        context={})


def remove_systemd_dropins():
    if not os.path.exists(SYSTEMD_LXD_DROP_IN_FOLDER):
        return

    shutil.rmtree(SYSTEMD_LXD_DROP_IN_FOLDER)


@when_any('config.changed.nagios_context',
          'config.changed.nagios_servicegroups')
def nagios_config_changed():
    if not is_flag_set('endpoint.nrpe-external-master.joined'):
        return
    update_nrpe_config()


@when('nrpe-external-master.available')
@when_not('nagios.configured')
def update_nrpe_config():
    hostname = nrpe.get_nagios_hostname()
    address = get_host_ip(unit_get('private-address'))
    port = config('https_port')
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    cmd = [
        '/usr/lib/nagios/plugins/check_http',
        '--ssl', '--hostname', "{}:{}".format(address, port),
        '--url', '/1.0', '--expect', '200',
        '--warning', '5', '--critical', '10'
    ]
    nrpe_setup.add_check(
        shortname='lxd_https',
        description='Check LXD https endpoint',
        check_cmd=" ".join(cmd),
    )
    nrpe_setup.write()
    set_flag('nagios.configured')
