#
# Anbox - The Android in a Box runtime environment
# Copyright 2020 Canonical Ltd.  All rights reserved.
#

import json
import os
import re
import urllib.request
from subprocess import check_call, check_output, CalledProcessError

from charmhelpers import fetch
from charmhelpers.core import hookenv
from charmhelpers.core.hookenv import status_set, config, local_unit, log
from charmhelpers.core.unitdata import kv
from charms.reactive import hook, when_any, when
from charms.reactive.flags import set_flag, clear_flag, is_flag_set

UA_STATUS_PATH = "/var/lib/ubuntu-advantage/status.json"
UA_MACHINE_TOKEN_PATH = "/var/lib/ubuntu-advantage/private/machine-token.json"

CONTRACTS_SERVICE_STAGING_URL = "https://contracts.staging.canonical.com"

UA_CLIENT_PREMIUM_PPA = "ppa:canonical-server/ua-client-premium"

unit_name = local_unit().replace("/", "-")
ANBOX_CLOUD_APT_SOURCES_FILE = "/etc/apt/sources.list.d/{}-anbox-cloud.list".format(unit_name)
ANBOX_CLOUD_APT_AUTH_FILE = "/etc/apt/auth.conf.d/90{}-anbox-cloud".format(unit_name)
ANBOX_CLOUD_APT_KEY_FILE = '/etc/apt/trusted.gpg.d/{}-anbox-cloud.asc'.format(unit_name)

UBUNTU_KEYSERVER_URL = 'https://keyserver.ubuntu.com'

ANBOX_CLOUD_STABLE_ARCHIVE_KEY_ID = "0x6B787E5FB493FA4FD0095D106DEC2758ACD5A973"

"""
Anbox Cloud UA Layer

Grants access to Anbox Cloud private archive using either a UA token, or
custom source and key.

The custom source will always take precedence over the UA token, however the token
will always be attached if available.
If ua_token is removed, a `ua detach` is performed.
If both the ua_token and the custom sources are removed, the archives are removed as well.
Changing `ua_token` to an invalid one will perform a detach on the machine but the
sources will remain configured.

The `ua.ready` flag is set as long as at least one access method is configured.
The `ua.changed` flag is set when the archive has changed. APT will be updated by that
point so there's no need to do it again .

The `ua.changed` flag will only be set if a new configuration is valid. e.g. changing the
ua_source to an invalid one will not update the archive and will not set `ua.changed`

On stop, a detach is performed, and the archives are removed if this charm is the one
that attached the token
"""

"""
Flags:
Public:
 - ua.ready: PPA have been configured and debian packages can be fetched
 - ua.changed: PPA sources have changed, a package upgrade might be required. Must be cleared manually

Private:
 - ua.attached: A UA token has been registered
 - ppa.changed: PPA configuration has changed, whether from UA or a custom source. It hasn't been
                written yet and only lives in the charm kv()
"""


def ensure_ua_client():
    try:
        raw_version = str(check_output(['ua', 'version']), 'utf-8')
        log('UA client version is {}'.format(raw_version), level='INFO')
        version = int(raw_version.split('.')[0])
        if version >= 20:
            return
    except Exception as err:
        log('Failed to determine UA client version: {}'.format(err), level='ERROR')

    fetch.apt_install(['software-properties-common'], fatal=True)

    try:
        check_call(['add-apt-repository', UA_CLIENT_PREMIUM_PPA])
    except CalledProcessError as err:
        raise Exception("Failed to setup UA PPA: {}".format(err))

    fetch.apt_update(fatal=False)
    fetch.apt_install(['ubuntu-advantage-tools'], fatal=True)


@hook('stop')
def stop():
    if not is_attached_by_other_application():
        try_ua_detach()
    kv().unset('ppa_ua_source')
    kv().unset('ppa_custom_source')
    clear_archive()


@when('ppa.changed')
def ppa_changed():
    clear_flag('ppa.changed')
    cfg = kv().get('ppa_custom_source', record=True) or kv().get('ppa_ua_source', record=True)
    if cfg is None:
        clear_flag('ua.ready')
        status_set('blocked', 'UA token missing')
        clear_archive()
        return

    add_ppa(url=cfg.url, login=cfg.login, password=cfg.password, ppa_key=cfg.ppa_key)
    fetch.apt_update()

    set_flag('ua.ready')
    set_flag('ua.changed')


@when_any('config.changed.ua_source', 'config.changed.ua_source_key', 'start')
def custom_source_config_changed():
    ua_source = config('ua_source')
    ua_source_key = config('ua_source_key')

    # We only remove custom ppa configuration when no item is set
    if not ua_source and not ua_source_key:
        kv().unset('ppa_custom_source')
        set_flag('ppa.changed')
        return

    if len(ua_source) == 0 and len(ua_source_key) > 0:
        status_set('blocked', 'ua_source must be set if ua_source_key is not empty')
        return
    elif len(ua_source) > 0 and len(ua_source_key) == 0:
        status_set('blocked', 'ua_source_key must be set if ua_source is not empty')
        return

    # Sanitize inputs if they are read from a file
    ua_source = ua_source.strip()
    ua_source_key = ua_source_key.strip()

    ppa_info = ua_source.split()
    if len(ppa_info) < 4:
        status_set('blocked', "malformed ua_source")
        return

    ppa_url = ppa_info[1]
    ppa_series = ppa_info[2]

    ensure_series_is_supported(get_current_series(), ppa_series)

    # Remove credentials from url and store them securely elsewhere
    creds = re.search('(?<=//).*(?=@)', ppa_url)
    if not creds or not creds.group(0):
        status_set('blocked', 'invalid ua_source: missing authentication')
        return

    creds = creds.group(0)
    safe_url = ppa_url.replace('{}@'.format(creds), '')

    creds = creds.split(':')
    if len(creds) != 2:
        status_set('blocked', 'invalid ua_source: malformed authentication')
        return
    username = creds[0]
    password = creds[1]

    public_key = None
    if ua_source_key == ANBOX_CLOUD_STABLE_ARCHIVE_KEY_ID:
        log('Taking stable archive key from the charm for {}'.format(ua_source_key), level='DEBUG')
        with open('files/anbox-cloud.pem') as f:
            public_key = f.read()
    else:
        if not ua_source_key.startswith('0x'):
            ua_source_key = '0x{}'.format(ua_source_key)
        search_url = '{}/pks/lookup?op=get&search={}'.format(UBUNTU_KEYSERVER_URL, ua_source_key)
        log('Fetching key with fingerprint {} at {}'.format(ua_source_key, search_url), level='DEBUG')
        public_key = urllib.request.urlopen(search_url, timeout=300).read().decode("utf-8")

    if public_key is None:
        raise Exception('Missing public key for UA source')

    new_ppa_info = {
        'url': safe_url,
        'login': username,
        'password': password,
        'ppa_key': public_key
    }
    # Don't update PPAs if nothing has changed
    if kv().delta(new_ppa_info, 'ppa_custom_source'):
        kv().set('ppa_custom_source', new_ppa_info)
        set_flag('ppa.changed')


@when_any('config.changed.ua_token', 'config.changed.ua_use_staging', 'start')
def ua_config_changed():
    try:
        try_ua_attach()
    except Exception as e:
        log('Failed to attach: {}'.format(e), level='ERROR')
        raise

    # ua_token was unset
    if not os.path.exists(UA_MACHINE_TOKEN_PATH):
        return

    with open(UA_MACHINE_TOKEN_PATH) as machine_token_file:
        data = json.load(machine_token_file)
        has_anbox_cloud = False
        for resource in data['availableResources']:
            if resource['name'] == "anbox-cloud" and resource['available'] is True:
                has_anbox_cloud = True
                break

        if not has_anbox_cloud:
            status_set('blocked', 'UA token is not entitled to Anbox Cloud')
            return

        contract_info = data['machineTokenInfo']['contractInfo']
        ent_by_name = dict((e["type"], e) for e in contract_info["resourceEntitlements"])

        ent = ent_by_name['anbox-cloud']

        supported_series = ent['affordances']['series']
        current_series = get_current_series()
        ensure_series_is_supported(current_series, supported_series)

        supported_architectures = ent['affordances']['architectures']
        ensure_architecture_is_supported(supported_architectures)

        tokens_by_name = dict((ent["type"], ent["token"]) for ent in data.get("resourceTokens", []))
        apt_resource_token = tokens_by_name['anbox-cloud']
        apt_url = ent['directives']['aptURL']

        public_key = None
        with open('files/anbox-cloud.pem') as f:
            public_key = f.read()
        if public_key is None:
            raise Exception('Failed to read archive public key')

        new_ppa_info = {
            'url': apt_url,
            'login': 'bearer',
            'password': apt_resource_token,
            'ppa_key': public_key
        }
        # Don't update PPAs if nothing has changed
        if kv().delta(new_ppa_info, 'ppa_ua_source'):
            kv().set('ppa_ua_source', new_ppa_info)
            set_flag('ppa.changed')


def load_image_server_resource_token():
    with open(UA_MACHINE_TOKEN_PATH) as machine_token_file:
        data = json.load(machine_token_file)
        tokens_by_name = dict((ent["type"], ent["token"]) for ent in data.get("resourceTokens", []))
        return tokens_by_name['anbox-images']


def is_attached_by_other_application():
    attached = True
    if not os.path.exists(UA_STATUS_PATH):
        attached = False
    else:
        with open(UA_STATUS_PATH, 'r') as f:
            status = json.loads(f.read())
            attached = status['attached']

    if attached and not is_flag_set('ua.attached'):
        return True

    return False


def try_ua_attach():
    if is_attached_by_other_application():
        log('UA already configured by other application', level='WARNING')
        return

    # We can't attach if we are already attached
    try_ua_detach()

    ua_token = config('ua_token')
    if not ua_token:
        kv().unset('ppa_ua_source')
        set_flag('ppa.changed')
        return

    # Sanitize inputs if they are read from a file
    ua_token = ua_token.strip()

    args = ['ua', 'attach', ua_token]
    use_staging = config('ua_use_staging') or False  # avoid None
    log('Attempting to attach UA token (staging: {})'.format(use_staging), level='INFO')
    if use_staging:
        args = ['env', 'UA_CONTRACT_URL={}'.format(CONTRACTS_SERVICE_STAGING_URL)] + args

    check_call(args)

    # If ua_use_staging config option is changed, we won't be able to
    # retrieve the old version, so we keep track of it in kv()
    kv().set('ua_attached_with_staging', '{}'.format(use_staging))
    set_flag('ua.attached')


def try_ua_detach():
    """Attempt to detach the UA token if already attached. Otherwise does nothing"""
    attached = True
    if not os.path.exists(UA_STATUS_PATH):
        attached = False
    else:
        with open(UA_STATUS_PATH, 'r') as f:
            status = json.loads(f.read())
            attached = status['attached']

    if not attached:
        log('UA not attached, not attempting to detach', level='INFO')
        return

    using_staging = kv().get('ua_attached_with_staging') == 'True'
    log('Attempting to detach UA token (staging: {})'.format(using_staging), level='INFO')
    args = ['ua', 'detach', '--assume-yes']
    if using_staging:
        args = ['env', 'UA_CONTRACT_URL={}'.format(CONTRACTS_SERVICE_STAGING_URL)] + args
    check_call(args)


def add_ppa(url: str, login: str, password: str, ppa_key: str):
    """
    Add a PPA to the system sources and updates apt cache. Overrides previous
    PPA if it exists
    :param: url (str) URL to the PPA. It should not contain the authentication token
    :param: login (str) PPA username
    :param: password (str) PPA password
    :param: ppa_key (str) PPA public key
    """
    with open(ANBOX_CLOUD_APT_SOURCES_FILE, 'w+') as f:
        f.write("deb {} {} main\n".format(url, get_current_series()))

    # APT on bionic can't deal with URLs in /etc/apt/auth.conf.d which have a protocol set
    # See apt_auth.conf(5)
    if login and password:
        auth_url = url.replace('https://', '')
        with open(os.open(ANBOX_CLOUD_APT_AUTH_FILE, os.O_CREAT | os.O_WRONLY, 0o600), 'w+') as f:
            f.write("machine {} login {} password {} # Anbox Cloud".format(auth_url, login, password))

    if ppa_key:
        with open(os.open(ANBOX_CLOUD_APT_KEY_FILE, os.O_CREAT | os.O_WRONLY, 0o644), 'w+') as f:
            f.write(ppa_key)

    fetch.apt_update(fatal=True)


def clear_archive():
    for f in [ANBOX_CLOUD_APT_SOURCES_FILE, ANBOX_CLOUD_APT_AUTH_FILE, ANBOX_CLOUD_APT_KEY_FILE]:
        if os.path.exists(f):
            os.remove(f)
    fetch.apt_update(fatal=False)


def get_current_series():
    return str(check_output(['lsb_release', '-c', '-s']), 'utf-8').strip('\n')


def ensure_series_is_supported(current_series, supported_series):
    if current_series not in supported_series:
        raise Exception("Current Ubuntu series {} is unsupported".format(current_series))


def ensure_architecture_is_supported(supported_architectures):
    current_arch = str(check_output(['uname', '-m']), 'utf-8').strip('\n')
    if current_arch == "aarch64":
        current_arch = "arm64"
    if current_arch not in supported_architectures:
        raise Exception("Current architecture {} is unsupported".format(
            current_arch))


hookenv.atstart(ensure_ua_client)
