#!/bin/sh -ex
#
# Anbox - The Android in a Box runtime environment
# Copyright 2018 Canonical Ltd.  All rights reserved.
#

distro=
arch=
# We track the current feature branch instead of the production version (see
# https://www.nvidia.com/en-us/drivers/unix/) because newer versions of the
# Video Codec SDK (needed for arm64 support) require >= 455.
driver_version=470
use_server_package=true
cuda_repo_pub_key=7fa2af80
nvidia_ubuntu_archive_pin_file=/etc/apt/preferences.d/nvidia-ubuntu-archive-pin-900
nvidia_udev_rule_file=/lib/udev/rules.d/71-anbox-cloud-nvidia.rules
nvidia_modprobe_conf_file=/etc/modprobe.d/anbox-nvidia.conf
anbox_nvidia_loader=/usr/local/bin/anbox-nvidia-loader

codename=$(lsb_release -c -s)
case "$codename" in
    bionic)
        distro=ubuntu1804
        ;;
    focal)
        distro=ubuntu2004
        ;;
    *)
        echo "ERROR: Unsupported Ubuntu release $codename"
        exit 1
        ;;
esac

hostarch=$(uname -m)
case "$hostarch" in
    aarch64)
        arch=sbsa
        ;;
    x86_64)
        arch=x86_64
        ;;
    *)
        echo "ERROR: Unsupport architecture $hostarch"
        exit 1
        ;;
esac

installed_nvidia_driver_version() {
  echo $(nvidia-smi | grep -o 'Driver Version: .*' | cut -d' ' -f3 | cut -d'.' -f1)
}

kernel_version() {
  echo $(uname -r)
}

apt_retry() {
  local retries=10
  local i=0
  # Wait for any concurrent apt processes to finish their work
  # before updating and installing
  while [ "$i" -lt "$retries" ]; do
    if [ "$i" -eq "$retries" ]; then
      echo "Failed to install $@ after $retries tries"
      exit 1
    fi

    if apt $@; then
      break
    fi

    sleep 5

    i=$((i+1))
  done
}

install_nvidia_drivers() {
    local version=$1
    local prebuilt_kernel_modules=$2
    local server_pkg=$3

    if [ "$arch" = x86_64 ]; then
        local pkg_opt=
        if [ "$server_pkg" = true ]; then
            pkg_opt="-server"
        fi

        local args=
        if [ "$prebuilt_kernel_modules" = true ]; then
            args=linux-modules-nvidia-"${version}${pkg_opt}"-"$(kernel_version)"
        else
            args=nvidia-headless-"${version}${pkg_opt}"
        fi

        # Since 1.10 Anbox Cloud release, except the prebuilt kernel module,
        # all others nvidia drivers packages are server oriented ones.
        apt_retry install -y --no-install-recommends \
              nvidia-utils-"${version}${pkg_opt}" \
              libnvidia-cfg1-"${version}${pkg_opt}" \
              libnvidia-common-"${version}${pkg_opt}" \
              libnvidia-compute-"${version}${pkg_opt}" \
              libnvidia-decode-"${version}${pkg_opt}" \
              libnvidia-encode-"${version}${pkg_opt}" \
              libnvidia-extra-"${version}${pkg_opt}" \
              libnvidia-fbc1-"${version}${pkg_opt}" \
              libnvidia-gl-"${version}${pkg_opt}" \
              libnvidia-ifr1-"${version}${pkg_opt}" $args
    else
        apt_retry install -y --no-install-recommends \
              nvidia-headless-"$version" \
              nvidia-utils-"$version" \
              libnvidia-cfg1-"$version" \
              libnvidia-common-"$version" \
              libnvidia-compute-"$version" \
              libnvidia-decode-"$version" \
              libnvidia-encode-"$version" \
              libnvidia-extra-"$version" \
              libnvidia-fbc1-"$version" \
              libnvidia-gl-"$version" \
              libnvidia-ifr1-"$version"
    fi
}

uninstall_nvidia_drivers() {
    local version=$1
    if [ "$arch" = x86_64 ]; then
        # Since 1.10 Anbox Cloud release, except the prebuilt kernel
        # module and cuda libraries packages, all others nvidia drivers
        # packages are server oriented ones. In order to remove old non
        # server packages, we append '*' in each package name to include
        # both package patterns.
        apt_retry remove --purge -y \
            nvidia-utils-"$version"* \
            libnvidia-cfg1-"$version"* \
            libnvidia-common-"$version"* \
            libnvidia-compute-"$version"* \
            libnvidia-decode-"$version"* \
            libnvidia-encode-"$version"* \
            libnvidia-extra-"$version"* \
            libnvidia-fbc1-"$version"* \
            libnvidia-gl-"$version"* \
            libnvidia-ifr1-"$version"* \
            linux-modules-nvidia-"$version"-"$(kernel_version)" \
            nvidia-headless-"$version"* \
            nvidia-dkms-"$version"* \
            nvidia-kernel-*-"$version"*
    else
        apt_retry remove --purge -y \
            nvidia-headless-"$version"* \
            nvidia-utils-"$version"* \
            libnvidia-cfg1-"$version"* \
            libnvidia-common-"$version"* \
            libnvidia-compute-"$version"* \
            libnvidia-decode-"$version"* \
            libnvidia-encode-"$version"* \
            libnvidia-extra-"$version"* \
            libnvidia-fbc1-"$version"* \
            libnvidia-gl-"$version"* \
            libnvidia-ifr1-"$version"* \
            nvidia-dkms-"$version"* \
            nvidia-kernel-*-"$version"*
    fi

    uninstall_cuda_libraries_when_needed
}

uninstall_cuda_libraries_when_needed() {
    if apt -qq list  | grep cuda-libraries- | grep installed ; then
        apt_retry remove --purge -y \
            cuda-libraries-* \
            cuda-*
    fi
}

load_nvidia_kernel_modules() {
    for module in nvidia_uvm nvidia_drm nvidia_modeset nvidia; do
       modprobe $module || true
    done
}

remove_nvidia_kernel_modules() {
    for module in nvidia_uvm nvidia_drm nvidia_modeset nvidia; do
       rmmod $module || true
    done
}

generate_pin_file_when_needed() {
    if [ "$arch" != x86_64 ]; then
        return
    fi

    # For x86_64, always get nvidia drivers from the ubuntu archive,
    # Hence, bump up the priority of nvidia drivers from the ubuntu archive
    cat << EOF > "$nvidia_ubuntu_archive_pin_file"
Package: nvidia*
Pin: origin *ubuntu.com*
Pin-Priority: 900

Package: libnvidia*
Pin: origin *ubuntu.com*
Pin-Priority: 900
EOF
}

import_cuda_repo_key_when_needed() {
  # We install nvidia drivers only from cuda repo for ARM64
  # and always install them from ubuntu archive on x86_64.
  if [ "$arch" = x86_64 ]; then
    return
  fi

  # See https://developer.nvidia.com/cuda-downloads for details
  apt-key adv --fetch-keys https://developer.download.nvidia.com/compute/cuda/repos/"$distro"/"$arch"/"$cuda_repo_pub_key".pub
  add-apt-repository "deb https://developer.download.nvidia.com/compute/cuda/repos/$distro/$arch/ /"

  wget https://developer.download.nvidia.com/compute/cuda/repos/"$distro"/"$arch"/cuda-"$distro".pin
  mv cuda-"$distro".pin /etc/apt/preferences.d/cuda-repository-pin-600
}

remove_cuda_repo_when_needed() {
  # Since Anbox Cloud 1.10 release, we don't need to install cuda libraries anymore
  # as Anbox Cloud doesn't require it to operate. Also on x86_64, all nvidia drivers
  # are installed from ubuntu archive, hence we need to drop the cuda repo and delete
  # the imported repo key for x86_64 as well.
  if [ "$arch" != x86_64 ]; then
    return
  fi

  sed -i '/developer.download.nvidia.com/d' /etc/apt/sources.list
  if apt-key list | grep  -i "nvidia.com" ; then
    apt-key del "$cuda_repo_pub_key"
  fi
}

install_udev_rule_when_needed() {
  # 1. For x86_64, installing the prebuilt kernel modules causes the nvidia kernel
  # modules are created are not created successfully after the drivers are installed.
  # This is because the udev file is missing and it's only available in
  # `nvidia-kernel-common-<version>` package which requires to build kernel modules
  # from the source. Hence we have to put anbox specific udev rule file in place.
  # 2. For ARM64, the udev rule file is missing which only causes the nvidia-uvm module
  # is not loaded properly on system reboot and the regarding device node is
  # not created sucessfully either. For this reason, we put the following udev
  # rule in place to load kernel module and create device nodes on system boot.
  if [ "$arch" != x86_64 ]; then
    cat << EOF > $nvidia_udev_rule_file
# Load and unload nvidia-uvm module
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe nvidia-uvm"
ACTION=="remove", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe -r nvidia-uvm"

# Create the device node for the nvidia-uvm module
ACTION=="add", DEVPATH=="/module/nvidia_uvm", SUBSYSTEM=="module", RUN+="$anbox_nvidia_loader"
EOF
  elif [ ! -e "/lib/udev/rules.d/71-nvidia.rules" ]; then
    cat << EOF > $nvidia_udev_rule_file
# Tag the device as master-of-seat so that logind is happy
# (see LP: #1365336)
SUBSYSTEM=="pci", ATTRS{vendor}=="0x10de", DRIVERS=="nvidia", TAG+="seat", TAG+="master-of-seat"

# Start and stop nvidia-persistenced on power on and power off
# respectively
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", TAG+="systemd", ENV{SYSTEMD_WANTS}="nvidia-persistenced.service"

# Load and unload nvidia-modeset module
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe nvidia-modeset"
ACTION=="remove", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe -r nvidia-modeset"

# Load and unload nvidia-drm module
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe nvidia-drm"
ACTION=="remove", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe -r nvidia-drm"

# Load and unload nvidia-uvm module
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe nvidia-uvm"
ACTION=="remove", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="/sbin/modprobe -r nvidia-uvm"

# This will create the device nvidia device nodes
ACTION=="add", DEVPATH=="/bus/pci/drivers/nvidia", RUN+="$anbox_nvidia_loader"

# Create the device node for the nvidia-uvm module
ACTION=="add", DEVPATH=="/module/nvidia_uvm", SUBSYSTEM=="module", RUN+="$anbox_nvidia_loader"

# Enable runtime PM for NVIDIA VGA/3D controller devices
ACTION=="add", SUBSYSTEM=="pci", ATTR{vendor}=="0x10de", ATTR{class}=="0x03[0-9]*", TEST=="power/control", ATTR{power/control}="auto"

# Enable runtime PM for NVIDIA Audio devices
ACTION=="add", SUBSYSTEM=="pci", ATTR{vendor}=="0x10de", ATTR{class}=="0x040300", TEST=="power/control", ATTR{power/control}="auto"

# Enable runtime PM for NVIDIA USB xHCI Host Controller devices
ACTION=="add", SUBSYSTEM=="pci", ATTR{vendor}=="0x10de", ATTR{class}=="0x0c0330", TEST=="power/control", ATTR{power/control}="auto"

# Enable runtime PM fo NVIDIA USB Type-C UCSI devices
ACTION=="add", SUBSYSTEM=="pci", ATTR{vendor}=="0x10de", ATTR{class}=="0x0c8000", TEST=="power/control", ATTR{power/control}="auto"
EOF
  fi

  # The ub-device-create tool is only available on x86_64 from the
  # nvidia-kernel-common package. For arm64, we create it by ourselves.
  cat <<'EOF'> "$anbox_nvidia_loader"
#!/bin/sh
# In case the kernel module is no loaded by udev rule
modprobe nvidia-uvm

# Find out the major device number used by the nvidia-uvm driver
rm -f /dev/nvidia-uvm
major_num=$(grep nvidia-uvm /proc/devices | awk '{print $1}')
mknod -m 666 /dev/nvidia-uvm c $major_num 0

nums_of_3d_controller=$(lspci | grep -i NVIDIA | grep "3D controller" | wc -l)
nums_of_vga=$(lspci | grep -i NVIDIA | grep "VGA compatible controller" | wc -l)
count=$(($nums_of_3d_controller + $nums_of_vga - 1))
for i in `seq 0 $count`; do
  rm -f /dev/nvidia$i
  mknod -m 666 /dev/nvidia$i c 195 $i;
done

rm -f /dev/nvidiactl
mknod -m 666 /dev/nvidiactl c 195 255
EOF
  chmod 0755 "$anbox_nvidia_loader"
}

install_settings_service() {
  cat << EOF > /usr/local/bin/anbox-nvidia-settings
#!/bin/bash -ex
# Enable persistence mode
nvidia-smi -pm 1

# Bump default clocks to avoid auto ramp up. nvidia-smi sorts the clock speeds
# from max to min already so we can easily determine the max speed
while read -r line ; do
  idx=\$(echo "\$line" | cut -d',' -f1)
  max_graphics_clock=\$(echo "\$line" | cut -d',' -f2 | xargs | cut -d' ' -f1)
  max_mem_clock=\$(echo "\$line" | cut -d',' -f3 | xargs | cut -d' ' -f1)
  # Some GPUs do not report any clock speeds
  if [ "\$max_graphics_clock" != "[N/A]" ]  && [ "\$max_mem_clock" != "[N/A]" ]; then
    nvidia-smi -i "\$idx" -ac "\$max_mem_clock","\$max_graphics_clock"
    nvidia-smi -i "\$idx" -lgc "\$max_graphics_clock","\$max_graphics_clock"
  fi
  ecc_mode=\$(echo "\$line" | cut -d',' -f4 | xargs | cut -d' ' -f1)
  if [ "\$ecc_mode" != "[N/A]" ]; then
    # Disable ECC memory, can give a 5% boost but requires a reboot
    nvidia-smi -i "\$idx" -e 0
  fi
done <<< \$(nvidia-smi --query-gpu=index,clocks.max.graphics,clocks.max.memory,ecc.mode.current --format=csv,noheader)
EOF
  chmod +x /usr/local/bin/anbox-nvidia-settings
  cat << EOF > /etc/systemd/system/anbox-nvidia-settings.service
[Unit]
Description=Apply Anbox specific Nvidia settings
Requires=nvidia-persistenced.service
After=nvidia-persistenced.service
[Service]
ExecStart=/usr/local/bin/anbox-nvidia-settings
Type=oneshot
[Install]
WantedBy=multi-user.target
# We need to be started again when the nvidia-persistenced service is
# restarted to ensure our settings are applied
WantedBy=nvidia-persistenced.service
EOF
  systemctl daemon-reload
  # Ensure that persistenced is started as otherwise it will override our settings
  systemctl start nvidia-persistenced.service
  systemctl enable --now anbox-nvidia-settings
}

uninstall_settings_service() {
  if [ -e /etc/systemd/system/anbox-nvidia-settings.service ]; then
    systemctl disable anbox-nvidia-settings
    rm /etc/systemd/system/anbox-nvidia-settings.service
    systemctl daemon-reload
  fi
  if [ -e /usr/local/bin/anbox-nvidia-settings ]; then
    rm /usr/local/bin/anbox-nvidia-settings
  fi
}

forbid_unattended_upgrade_when_needed() {
  # On ARM64, to avoid all nvidia packages get updated by unattended
  # upgrade from ubuntu archive and cause the conflict/veresion mismatch
  # from the one installed from cuda repo, we don't allow all nvidia
  # packages can be updated from ubuntu archive with unattended upgrade
  if [ "$arch" = x86_64 ]; then
    return
  fi

  unattended_upgrades_file=/etc/apt/apt.conf.d/50unattended-upgrades
  for pattern in libnvidia- nvidia- cuda-libraries- ; do
    if ! grep -oPz  '(?s)Unattended-Upgrade::Package-Blacklist.*?};' "$unattended_upgrades_file" \
      | grep "  $pattern" ; then
      sed -i "/^Unattended-Upgrade::Package-Blacklist/a\  $pattern;" "$unattended_upgrades_file"
    fi
  done
}

disable_nouveau_driver_when_needed() {
  if lsmod | grep nouveau; then
    echo "blacklist nouveau" > "$nvidia_modprobe_conf_file"
    echo "options nouveau modeset=0" >> "$nvidia_modprobe_conf_file"
    rmmod nouveau || true
    update-initramfs -u
  fi
}

configure_nvidia_kmod_if_needed() {
  # We have to bump rsvd memory size to allow higher number of simulatenous users
  # on the GPU from an API client perspective
  if ! cat "$nvidia_modprobe_conf_file" | grep -q 'RMIncreaseRsvdMemorySizeMB=2048'; then
    echo "options nvidia NVreg_RegistryDwords=RMIncreaseRsvdMemorySizeMB=2048" >> \
      "$nvidia_modprobe_conf_file"
  fi
}

nvidia_install_driver() {
    local prebuilt_kernel_modules=$1

    disable_nouveau_driver_when_needed

    configure_nvidia_kmod_if_needed

    import_cuda_repo_key_when_needed

    generate_pin_file_when_needed

    apt_retry update

    install_nvidia_drivers "$driver_version" "$prebuilt_kernel_modules" "$use_server_package"

    install_udev_rule_when_needed

    forbid_unattended_upgrade_when_needed

    # Load nvidia prebuilt kernel modules and create device nodes immediately
    load_nvidia_kernel_modules
    if [ -e "$anbox_nvidia_loader" ]; then
        $anbox_nvidia_loader
    fi

    # Ensure the Nvidia driver is functional. Calling nvidia-smi not as root will
    # prevent it from loading any missing drivers
    sudo -u ubuntu nvidia-smi

    install_settings_service
}

nvidia_switch_driver() {
    # The switch command only works on X86_64 based host where nvidia drivers
    # were installed. If the nvidia drivers are not installed,  people need
    # to run `gpu-support.sh nvidia install` first to install nvidia drivers
    # since beside the driver packages, there are other bits which are
    # going to be installed along the way on installation, E.g. udev rule.
    if [ "$arch" != x86_64 ]; then
      echo "The 'switch' command for nvidia GPU is only supported for x86_64 based host."
      exit 1
    fi

    if ! apt -qq list | grep nvidia-utils-"$driver_version"; then
      echo "Nvidia drivers are not found, please install nvidia drivers first."
      exit 1
    fi

    local switch_package=$1
    local server_pkg=false
    if [ "$switch_package" = "server" ]; then
      server_pkg=true
    fi

    # Check if prebuilt nvidia kernel modules are installed
    local prebuilt_kernel_modules=false
    local drivers_installed_version=$(installed_nvidia_driver_version)
    if apt -qq list | grep linux-modules-nvidia-"$drivers_installed_version" | grep installed ; then
        prebuilt_kernel_modules=true
    fi

    uninstall_nvidia_drivers "$drivers_installed_version"
    install_nvidia_drivers "$driver_version" "$prebuilt_kernel_modules" "$server_pkg"

    # Trigger load of needed kernel modules.
    remove_nvidia_kernel_modules
    nvidia-smi
}

nvidia_upgrade_driver() {
    local prebuilt_kernel_modules=$1

    if [ ! -e "$nvidia_ubuntu_archive_pin_file" ]; then
      generate_pin_file_when_needed
    fi

    if [ ! -e "$nvidia_udev_rule_file" ]; then
      install_udev_rule_when_needed
    fi

    configure_nvidia_kmod_if_needed

    # Always install the latest version of the settings service
    install_settings_service

    forbid_unattended_upgrade_when_needed

    # Anbox Cloud doesn't require `cuda-libraries` to operate so the package was
    # installed from the previous release, now it's time to get it completelty dropped.
    uninstall_cuda_libraries_when_needed

    remove_cuda_repo_when_needed

    apt_retry update

    local drivers_installed_version=$(installed_nvidia_driver_version)
    if [ $drivers_installed_version -lt $driver_version ]; then
        uninstall_nvidia_drivers "$drivers_installed_version"
        install_nvidia_drivers "$driver_version" "$prebuilt_kernel_modules" "$use_server_package"
    else
        # Check if server oriented packages are installed, if not
        # we migrate to server oriented packages first.
        if [ "$use_server_package" = true ] && ! apt list --installed | grep  "libnvidia-.*-$driver_version-server" ; then
            uninstall_nvidia_drivers "$driver_version"
            install_nvidia_drivers "$driver_version" "$prebuilt_kernel_modules" "$use_server_package"
        else
            apt_retry upgrade -y
            return 0
        fi
    fi

    # In case of hitting `Failed to initialize NVML: Driver/library version mismatch`
    # after upgrading the nvidia drivers, we have to manually remove nvidia kernel
    # modules before getting the new drivers loaded via `nvidia-smi`
    remove_nvidia_kernel_modules

    # Trigger load of needed kernel modules.
    nvidia-smi
}

nvidia_uninstall_driver() {
    uninstall_settings_service
    uninstall_nvidia_drivers "$driver_version"
    remove_nvidia_kernel_modules

    # Only attempt to remove leftover bits from nvidia-container-runtime installations
    # when it was installed by us before
    if [ -e /etc/apt/sources.list.d/nvidia-container-runtime.list ]; then
        rm -f /etc/apt/sources.list.d/nvidia-container-runtime.list
        apt_retry purge -y nvidia-container-runtime
    fi

    if [ -e "$nvidia_ubuntu_archive_pin_file" ]; then
        rm -f "$nvidia_ubuntu_archive_pin_file"
        apt_retry update
    fi

    if [ -e "$nvidia_udev_rule_file" ]; then
        rm -fr "$nvidia_udev_rule_file" "$anbox_nvidia_loader"
    fi

    if [ -e "$nvidia_modprobe_conf_file" ]; then
        rm -f "$nvidia_modprobe_conf_file"
    fi
}

run_nvidia_action() {
    # Within LXD we don't have to install the Nvidia driver as it is already
    # shared from the host and we don't maintain it
    if [ -e /dev/lxd ] && [ ! -e /dev/lxd/sock ] ; then
        echo "INFO: Running within LXD container, not installing Nvidia driver"
        exit 0
    fi

    local prebuilt_kernel_modules=false
    local switch_package=
    local action=$1; shift
    while [ -n "$1" ]; do
      case "$1" in
        --prebuilt-kernel-modules)
            prebuilt_kernel_modules=true
            shift
            ;;
        --package=*)
            switch_package=${1#*=}
            if [ "$switch_package" != "server" ] && [ "$switch_package" != "non-server" ] ; then
              echo "Invalid package usage, only 'server' and 'non-server' flags are supported"
              exit 1
            fi
            shift
            ;;
        *)
            echo "ERROR: unknown option $1"
            exit 1
            ;;
      esac
    done

    workdir=$(mktemp -d)
    cd "$workdir"

    case "$action" in
    install)
        nvidia_install_driver "$prebuilt_kernel_modules"
        ;;
    uninstall)
        nvidia_uninstall_driver
        ;;
    upgrade)
        nvidia_upgrade_driver "$prebuilt_kernel_modules"
        ;;
    switch)
        nvidia_switch_driver "$switch_package"
        ;;
    *)
        echo "ERROR: Invalid action"
        exit 1
    esac
}

GPU_TYPE="$1"
shift 1
ARGS=$@

case "$GPU_TYPE" in
    nvidia)
        run_nvidia_action $ARGS
        ;;
    intel)
        exit 0
        ;;
    amd)
        exit 0
        ;;
    drm)
        exit 0
        ;;
    none)
        exit 0
        ;;
    *)
        echo "ERROR: Invalid GPU type $GPU_TYPE"
        exit 1
        ;;
esac
